import sys
import numpy as np
from pydantic import BaseModel, validator


class GridConfig(BaseModel):
    seed: int = None
    size: int = 8
    density: float = 0.3
    num_agents: int = 1
    obs_radius: int = 5

    FREE = 0
    OBSTACLE = 1
    MOVES = tuple([(0, 0), (-1, 0), (1, 0), (0, -1), (0, 1),  ])

    # noinspection PyMethodParameters
    @validator('seed')
    def seed_initialization(cls, v):
        assert v is None or (0 <= v < sys.maxsize), "seed must be in [0, " + str(sys.maxsize) + ']'
        if v is None:
            return np.random.randint(sys.maxsize, dtype=np.int64)
        return v

    # noinspection PyMethodParameters
    @validator('size')
    def size_restrictions(cls, v):
        assert 2 <= v <= 1024, "size must be in [2, 1024]"
        return v

    # noinspection PyMethodParameters
    @validator('density')
    def density_restrictions(cls, v):
        assert 0.0 <= v < 1, "density must be in [0, 1)"
        return v

    # noinspection PyMethodParameters
    @validator('num_agents')
    def num_agents_must_be_positive(cls, v):
        assert 1 <= v <= 1024, "num_agents must be in [1, 1024]"
        return v

    # noinspection PyMethodParameters
    @validator('obs_radius')
    def obs_radius_must_be_positive(cls, v):
        assert 1 <= v <= 128, "obs_radius must be in [1, 128]"
        return v



import string
import sys
import time
from contextlib import closing

import numpy as np
from gym import utils
from io import StringIO

from pogema.generator import generate_obstacles, generate_positions_and_targets_fast
from .grid_config import GridConfig


def timeit(func):
    def wrapped(*args, **kwargs):
        start = time.monotonic()
        result = func(*args, **kwargs)
        end = time.monotonic()
        print(end - start)
        return result

    return wrapped


class Grid:

    def __init__(self, obstacles, starts, starts_xy, finishes_xy, grid_config: GridConfig):
        self.obstacles = obstacles
        self.positions = starts
        self.finishes_xy = finishes_xy
        self.positions_xy = starts_xy
        self.config = grid_config

    def get_observation_shape(self):
        full_radius = self.config.obs_radius * 2 + 1
        return 2, full_radius, full_radius

    def get_num_actions(self):
        return len(self.config.MOVES)

    def get_obstacles(self, agent_id):
        x, y = self.positions_xy[agent_id]
        r = self.config.obs_radius
        return self.obstacles[x - r:x + r + 1, y - r:y + r + 1]

    def get_positions(self, agent_id):
        x, y = self.positions_xy[agent_id]
        r = self.config.obs_radius
        return self.positions[x - r:x + r + 1, y - r:y + r + 1]

    def get_target(self, agent_id):

        x, y = self.positions_xy[agent_id]
        fx, fy = self.finishes_xy[agent_id]
        if x == fx and y == fy:
            return 0.0, 0.0
        rx, ry = fx - x, fy - y
        dist = np.sqrt(rx ** 2 + ry ** 2)
        return rx / dist, ry / dist

    def get_square_target(self, agent_id):
        c = self.config
        full_size = self.config.obs_radius * 2 + 1
        result = np.zeros((full_size, full_size))
        x, y = self.positions_xy[agent_id]
        fx, fy = self.finishes_xy[agent_id]
        dx, dy = x - fx, y - fy

        dx = min(dx, c.obs_radius) if dx >= 0 else max(dx, -c.obs_radius)
        dy = min(dy, c.obs_radius) if dy >= 0 else max(dy, -c.obs_radius)
        result[c.obs_radius + dx, c.obs_radius + dy] = 1
        return result

    def render(self, mode='human'):
        outfile = StringIO() if mode == 'ansi' else sys.stdout
        chars = string.digits + string.ascii_letters + string.punctuation
        positions_map = {(x, y): id_ for id_, (x, y) in enumerate(self.positions_xy)}
        finishes_map = {(x, y): id_ for id_, (x, y) in enumerate(self.finishes_xy)}
        for line_index, line in enumerate(self.obstacles):
            out = ''
            for cell_index, cell in enumerate(line):
                if cell == self.config.FREE:
                    agent_id = positions_map.get((line_index, cell_index), None)
                    finish_id = finishes_map.get((line_index, cell_index), None)

                    if agent_id is not None:
                        out += str(utils.colorize(' ' + chars[agent_id % len(chars)] + ' ', color='gray', bold=True,
                                                  highlight=False))
                    elif finish_id is not None:
                        out += str(
                            utils.colorize('|' + chars[finish_id % len(chars)] + '|', 'white', highlight=False))
                    else:
                        out += str(utils.colorize(str(' . '), 'white', highlight=False))
                else:
                    out += str(utils.colorize(str('   '), 'cyan', bold=False, highlight=True))
            out += '\n'
            outfile.write(out)

        if mode != 'human':
            with closing(outfile):
                return outfile.getvalue()

    def move(self, agent_id, action):
        x, y = self.positions_xy[agent_id]
        fx, fy = self.finishes_xy[agent_id]

        dx, dy = self.config.MOVES[action]
        if self.obstacles[x + dx, y + dy] == self.config.FREE:
            if self.positions[x + dx, y + dy] == self.config.FREE:
                x += dx
                y += dy
        if fx == x and fy == y:
            self.positions[x, y] = self.config.FREE
        self.positions_xy[agent_id] = (x, y)

        if x == fx and y == fy:
            return True

        return False

    @staticmethod
    def prepare_grid(obstacles, starts_xy, finishes_xy, grid_config):
        r = grid_config.obs_radius
        filled_obstacles = np.zeros(np.array(obstacles.shape) + r * 2)
        size, size = filled_obstacles.shape
        filled_obstacles[r - 1, r - 1:size - r + 1] = 1
        filled_obstacles[r - 1:size - r + 1, r - 1] = 1
        filled_obstacles[size - r, r - 1:size - r + 1] = 1
        filled_obstacles[r - 1:size - r + 1, size - r] = 1
        filled_obstacles[r:size - r, r:size - r] = obstacles

        starts_xy = [(x + r, y + r) for x, y in starts_xy]
        finishes_xy = [(x + r, y + r) for x, y in finishes_xy]
        filled_positions = np.zeros(filled_obstacles.shape)
        for x, y in starts_xy:
            filled_positions[x, y] = 1

        return Grid(filled_obstacles, filled_positions, starts_xy, finishes_xy, grid_config, )

    @staticmethod
    def random_grid_generator(config):
        obstacles = generate_obstacles(config)
        starts_xy, finishes_xy = generate_positions_and_targets_fast(obstacles, config)
        return Grid.prepare_grid(obstacles, starts_xy, finishes_xy, config)



from abc import ABC

import numpy as np
import gym
from gym.error import ResetNeeded

from pogema.generator import generate_obstacles
from pogema.grid import Grid
from pogema.grid_config import GridConfig


class PogemaBase(gym.Env):

    def __init__(self, config: GridConfig = GridConfig()):
        # noinspection PyTypeChecker
        self.grid: Grid = None
        self.config = config

        full_size = self.config.obs_radius * 2 + 1
        self.observation_space = gym.spaces.Box(0.0, 1.0, shape=(3, full_size, full_size))
        self.action_space = gym.spaces.Discrete(len(self.config.MOVES))

    def _get_agents_obs(self, agent_id=0):
        return np.concatenate([
            self.grid.get_obstacles(agent_id)[None],
            self.grid.get_positions(agent_id)[None],
            self.grid.get_square_target(agent_id)[None]
        ])

    def check_reset(self):
        if self.grid is None:
            raise ResetNeeded("Please reset environment first!")

    def render(self, mode='human'):
        self.check_reset()
        return self.grid.render(mode=mode)


class PogesaBase(PogemaBase):

    def __init__(self, config: GridConfig = GridConfig(num_agents=1)):
        # noinspection PyTypeChecker
        self.grid: Grid = None
        self.config = config

        full_size = self.config.obs_radius * 2 + 1
        self.observation_space = gym.spaces.Box(0.0, 1.0, shape=(3, full_size, full_size))
        self.action_space = gym.spaces.Discrete(len(self.config.MOVES))

    def step(self, action):
        raise NotImplemented

    # def _get_obs_dict(self, agent_id=0):
    #     observation = dict(
    #         obs=np.concatenate([
    #             self.grid.get_obstacles(agent_id)[None],
    #             self.grid.get_positions(agent_id)[None],
    #             self.grid.get_square_target(agent_id)[None]
    #         ]),
    #         target_vector=self.grid.get_target(agent_id)
    #     )
    #     return observation


class StochasticPogesa(PogesaBase):
    def __init__(self, config: GridConfig = GridConfig(num_agents=1), morph_tau=1):
        super().__init__(config)
        self.morph_tau = morph_tau
        self.rnd = np.random.default_rng(self.config.seed)
        self._steps = 0

    def morph_obstacles(self):
        r = self.config.obs_radius
        self.grid.obstacles[r:-r, r:-r] = generate_obstacles(self.config, self.rnd)
        for position in self.grid.positions_xy:
            self.grid.obstacles[position] = self.config.FREE

    def step(self, action):
        self.check_reset()

        if self._steps % self.morph_tau == 0:
            self.morph_obstacles()

        finish = self.grid.move(0, action)
        info = dict()
        reward = 1.0 if finish else 0.0

        self._steps += 1
        return self._get_obs(), reward, finish, info

    def reset(self):
        self._steps = 0
        self.grid: Grid = Grid.random_grid_generator(config=self.config)
        return self._get_obs()


class Pogesa(PogesaBase):

    def __init__(self, config: GridConfig = GridConfig(num_agents=1)):
        # noinspection PyTypeChecker
        super().__init__(config)

    def step(self, action):
        self.check_reset()

        finish = self.grid.move(0, action)
        info = dict()
        reward = 1.0 if finish else 0.0
        return self._get_obs(), reward, finish, info


class SingleAgentPogema(Pogesa):
    pass


class Pogema(PogemaBase):
    def __init__(self, config=GridConfig(num_agents=2)):
        super().__init__(config)
        # self.grid: Grid = None
        # self.config: GridConfig = config
        #
        # full_size = self.config.obs_radius * 2 + 1
        # self.observation_space = gym.spaces.Box(0.0, 1.0, shape=(3, full_size, full_size))
        # self.action_space = gym.spaces.Discrete(len(self.config.MOVES))

        self.num_agents = self.config.num_agents
        self.is_multiagent = True

    def _obs(self):
        return [self._get_agents_obs(index) for index in range(self.num_agents)]

    def step(self, action: list):
        assert len(action) == self.config.num_agents
        rewards = []

        infos = [dict() for _ in range(self.num_agents)]

        global_done = True

        for agent_idx in range(self.config.num_agents):
            agent_done = self.grid.move(agent_idx, action[agent_idx])
            infos[agent_idx]['is_active'] = not agent_done
            if not agent_done:
                global_done = False
            if agent_done:
                rewards.append(1.0)
            else:
                rewards.append(0.0)

        return self._obs(), rewards, global_done, infos

    def reset(self):
        self.grid: Grid = Grid.random_grid_generator(config=self.config)
        return self._obs()

    def render(self, mode="human"):
        pass


def main():
    env = gym.make('Pogema-v0', config=GridConfig(size=8, num_agents=2))
    obs = env.reset()

    for _ in range(1000):
        obs, reward, done, info = env.step([env.action_space.sample() for _ in range(env.config.num_agents)])
        print(obs, reward, done, info)
        env.render()
        if done:
            break


if __name__ == '__main__':
    main()



from collections import defaultdict
from itertools import cycle

import gym
import numpy as np
import drawSvg
from pathlib import Path

from pogema.grid_config import GridConfig

from pogema.grid import Grid


class MarlSvgVisualization:
    EMPTY = 0
    MOVES = [(1, 0), (-1, 0), (0, 1), (0, -1), ]
    RADIUS = 30

    SCALE_SIZE = 100
    CIRCLE_STROKEWIDTH = 10
    MOVING_CIRCLE_SIZE = 15
    ANIMATION_TIME_SCALE = 0.07 * 4
    OPACITY = 0.1
    COLORS = ('#ff5447', '#3F51B5', '#ecbb33',
              '#3d9e3e', '#da7424', '#c1433c', '#ade6ad', '#5cbbb9', '#9C27B0', '#5a349c')

    def __init__(self, obstacles, num_agents, targets, obs_radius, drawing_size=(500, 500)):
        """
        :param obstacles:
        :param num_agents:
        :param targets:
        :param obs_radius:
        :param drawing_size:
        """
        self.num_agents = num_agents
        self.targets = targets
        self.obstacles = obstacles
        self.obs_radius = obs_radius
        self.color = cycle(self.COLORS)
        self.settings = {"base_inner_color": '#7b8594', "base_outer_color": None, }
        self.agents_colors = {index: next(self.color) for index in range(self.num_agents)}
        self.paths = {index: [] for index in range(self.num_agents)}
        self.done_statuses = {index: [] for index in range(self.num_agents)}
        self.drawing_size = drawing_size

        self.START_I = 50 - self.obs_radius * self.SCALE_SIZE
        # self.START_I = 50

        self.START_J = 50 + self.obs_radius * self.SCALE_SIZE
        # self.START_J = 50

    def store(self, positions, done_statuses):
        """
        Stores agents positions and done statuses in episode buffer.
        :param positions: positions of agents
        :param done_statuses: done statuses of agents
        :return:
        """
        for agent_id, (x, y) in enumerate(positions):
            self.paths[agent_id].append((x, y))
        for agent_id, done_status in enumerate(done_statuses):
            self.done_statuses[agent_id].append(done_status)

    def draw(self):
        """
        Main function for visualization.
        Creates connections, nodes, targets and agents.
        :return: DrawSvg object
        """

        size = len(self.obstacles)

        drawing = drawSvg.Drawing(size * self.SCALE_SIZE, size * self.SCALE_SIZE, origin=(0, 0), displayInline=False)
        drawing.setRenderSize(*self.drawing_size)

        self.create_connections(drawing, size)
        self.create_nodes(drawing, size, self.settings, self.obstacles, self.paths, self.done_statuses, self.obs_radius)
        self.create_targets(drawing, self.targets, size, self.agents_colors, self.done_statuses)
        self.create_agents(drawing, self.paths, size, self.agents_colors, self.done_statuses)

        return drawing

    def check_connection(self, first_x, first_y, second_x, second_y, connection_id=0):
        """
        Checks connection between 2 nodes.
        :param first_x:
        :param first_y:
        :param second_x:
        :param second_y:
        :param connection_id: TODO I don't remember for what is this :)))
        :return:
        """
        board_size = len(self.obstacles)
        if not (0 <= first_x < board_size and 0 <= first_y < board_size):
            return False

        if not (0 <= second_x < board_size and 0 <= second_y < board_size):
            return False

        return self.obstacles[first_x][first_y] == connection_id and self.obstacles[second_x][second_y] == connection_id

    def create_connections(self, drawing, size):
        """
        Creates connections between nodes.
        Current implementations ignores existing connection (duplicates them).
        :param drawing: drawSvg object to which the nodes will be assigned
        :param size: board size
        :return:
        """
        for i in range(self.obs_radius, size - self.obs_radius):
            for j in range(self.obs_radius, size - self.obs_radius):
                for di, dj in self.MOVES:
                    if not self.check_connection(*self.fix_point(i, j, size), *self.fix_point(i + di, j + dj, size)):
                        continue
                    sx, sy = self.START_I + i * self.SCALE_SIZE, self.START_J + j * self.SCALE_SIZE
                    fx, fy = self.START_I + (i + di) * self.SCALE_SIZE, self.START_J + (j + dj) * self.SCALE_SIZE,

                    if fx < sx or fy < sy:
                        sx, fx = fx, sx
                        sy, fy = fy, sy
                    radius_offset = 0
                    if di:
                        element = drawSvg.Line(sx + self.RADIUS + radius_offset, sy, fx - self.RADIUS - radius_offset,
                                               fy, stroke='black', stroke_width=self.CIRCLE_STROKEWIDTH, fill='black',
                                               opacity=self.OPACITY)
                        drawing.append(element)
                    elif dj:
                        element = drawSvg.Line(sx, sy + self.RADIUS + radius_offset, fx,
                                               fy - self.RADIUS - radius_offset, stroke='black',
                                               stroke_width=self.CIRCLE_STROKEWIDTH, fill='black', opacity=self.OPACITY)
                        drawing.append(element)
                    else:
                        raise KeyError

    @staticmethod
    def fix_point(x, y, length):
        """
        Fixes coordinates to match their positions in arrays.
        :param x:
        :param y:
        :param length:
        :return: fixed coordinates pair
        """
        return length - y - 1, x

    def create_nodes(self, drawing, size, settings, board, paths, done_statuses, obs_radius):
        """
        Fully creates nodes and animation staff for agents observations.
        :param drawing: drawSvg object to which the nodes will be assigned
        :param size: board size
        :param settings: some settings TODO replace with self.attribute
        :param board:
        :param paths: agents paths
        :param done_statuses:
        :param obs_radius: agents observation radius
        :return:
        """
        nodes = defaultdict(lambda: [])

        for i in range(self.obs_radius, size - self.obs_radius):
            for j in range(self.obs_radius, size - self.obs_radius):
                x, y = self.fix_point(i, j, size)
                if board[x][y] == self.EMPTY:
                    # drawing inner circle
                    inner_color = None
                    if "base_inner_color" in settings:
                        inner_color = settings['base_inner_color']
                    inner_circle = drawSvg.Circle(self.START_I + i * self.SCALE_SIZE,
                                                  self.START_J + j * self.SCALE_SIZE,
                                                  self.RADIUS - self.CIRCLE_STROKEWIDTH // 2,
                                                  stroke='black', stroke_width=0, fill=inner_color,
                                                  opacity=self.OPACITY)
                    drawing.append(inner_circle)
                    # store nodes to apply opacity
                    nodes[x, y].append(inner_circle)

                    # drawing outer circle
                    outer_color = None
                    if "base_outer_color" in settings:
                        outer_color = settings['base_outer_color']

                    outer_circle = drawSvg.Circle(self.START_I + i * self.SCALE_SIZE,
                                                  self.START_J + j * self.SCALE_SIZE,
                                                  self.RADIUS + self.CIRCLE_STROKEWIDTH,
                                                  stroke=outer_color, stroke_width=self.CIRCLE_STROKEWIDTH, fill='none',
                                                  opacity=self.OPACITY)
                    drawing.append(outer_circle)
                    nodes[x, y].append(outer_circle)

                    # clearing connections with white circle
                    white_circle = drawSvg.Circle(self.START_I + i * self.SCALE_SIZE,
                                                  self.START_J + j * self.SCALE_SIZE,
                                                  self.RADIUS, stroke='white', stroke_width=self.CIRCLE_STROKEWIDTH,
                                                  fill='none', opacity=1.0)
                    drawing.append(white_circle)

                    # drawing black border circle
                    border_circle = drawSvg.Circle(self.START_I + i * self.SCALE_SIZE,
                                                   self.START_J + j * self.SCALE_SIZE,
                                                   self.RADIUS, stroke='black', stroke_width=self.CIRCLE_STROKEWIDTH,
                                                   fill='none', opacity=self.OPACITY)
                    drawing.append(border_circle)
                    nodes[x, y].append(border_circle)

        # change nodes opacity based on agents vision (obs_radius)
        animation_length = 0
        for agent_id in sorted(paths.keys()):
            animation_length = max(len(paths[agent_id]), animation_length)

        view_range = [np.zeros((size, size), dtype=np.uint8) for _ in range(animation_length)]
        for agent_id in sorted(paths.keys()):
            agent_xy = paths[agent_id]
            for index, (x, y) in enumerate(agent_xy):
                # skip highlighting for done agents
                if done_statuses[agent_id][index]:
                    break
                lower_x, upper_x = max(0, x - obs_radius), min(size, x + obs_radius + 1)
                lower_y, upper_y = max(0, y - obs_radius), min(size, y + obs_radius + 1)
                view_range[index][lower_x:upper_x, lower_y:upper_y] = 1

        for x, y in nodes:
            for element in nodes[x, y]:
                if view_range[0][x, y] != 0:
                    element.args['opacity'] = '1.0'
                if len(view_range) > 1:
                    p = [str(self.OPACITY) if view_range[index][x, y] == 0 else "1.0" for index in
                         range(len(view_range))]
                    p = ';'.join(p)

                    element.appendAnim(drawSvg.Animate('opacity', f'{animation_length * self.ANIMATION_TIME_SCALE}s', p,
                                                       repeatCount='indefinite'))

    def create_targets(self, drawing, targets, size, targets_colors, done_statuses):
        """
        Creates agent's targets, also adds animation to hide
        target after reaching.
        :param drawing: drawSvg object to which the nodes will be assigned
        :param targets:
        :param size:
        :param targets_colors:
        :param done_statuses:
        :return:
        """
        for agent_id in sorted(targets.keys()):
            tx, ty = targets[agent_id]

            # fix coordinates
            x, y = ty, size - tx - 1

            target = drawSvg.Circle(self.START_I + x * self.SCALE_SIZE, self.START_J + y * self.SCALE_SIZE,
                                    self.RADIUS + self.CIRCLE_STROKEWIDTH,
                                    stroke=targets_colors[agent_id], stroke_width=self.CIRCLE_STROKEWIDTH, fill='none')
            animation_length = len(done_statuses[agent_id]) * self.ANIMATION_TIME_SCALE
            visibility = ";".join(['hidden' if is_done else "visible" for is_done in done_statuses[agent_id]])
            if len(done_statuses[agent_id]) > 1:
                target.appendAnim(
                    drawSvg.Animate('visibility', f'{animation_length}s', visibility, repeatCount='indefinite'))
            drawing.append(target)

    def create_agents(self, drawing, paths, size, agents_colors, done_statuses):
        """
        Creates agents and their move animations, also hides
        agents after reaching a target.
        :param drawing: drawSvg object to which the nodes will be assigned
        :param paths:
        :param size:
        :param agents_colors:
        :param done_statuses:
        :return:
        """
        for agent_id in sorted(paths.keys()):
            # skip agent drawing if its done
            if self.done_statuses[agent_id][0] is True:
                continue

            agent_xy = paths[agent_id]

            pr = self.RADIUS - self.CIRCLE_STROKEWIDTH // 2
            agent = drawSvg.Circle(self.START_I + agent_xy[0][1] * self.SCALE_SIZE,
                                   self.START_J + (size - agent_xy[0][0] - 1) * self.SCALE_SIZE,
                                   pr, stroke='black', stroke_width=0,
                                   fill=agents_colors[agent_id])
            x_path = "".join([str(self.START_I + qy * self.SCALE_SIZE) + ";" for qx, qy in agent_xy])
            y_path = "".join([str(-self.START_J + -(size - qx - 1) * self.SCALE_SIZE) + ";" for qx, qy in agent_xy])
            if len(agent_xy) > 1:
                animation_length = len(agent_xy) * self.ANIMATION_TIME_SCALE
                agent.appendAnim(drawSvg.Animate('cy', f'{animation_length}s', y_path, repeatCount='indefinite'))
                agent.appendAnim(drawSvg.Animate('cx', f'{animation_length}s', x_path, repeatCount='indefinite'))

                visibility = ";".join(['hidden' if is_done else "visible" for is_done in done_statuses[agent_id]])
                agent.appendAnim(
                    drawSvg.Animate('visibility', f'{animation_length}s', visibility, repeatCount='indefinite'))
            drawing.append(agent)

    @staticmethod
    def get_frame(obstacles, agents, targets, obs_radius=5, done=None):
        targets = {agent_id: target for agent_id, target in enumerate(targets)}
        msv = MarlSvgVisualization(obstacles=obstacles, num_agents=len(agents), targets=targets, obs_radius=obs_radius)
        if done is None:
            done = [False for _ in range(len(agents))]
        msv.store(agents, done)
        return msv.draw()

    @staticmethod
    def draw_frame(obstacles, agents, targets, filename, obs_radius=5, done=None):
        Path(filename).parents[0].mkdir(parents=True, exist_ok=True)
        drawing = MarlSvgVisualization.get_frame(obstacles, agents, targets, obs_radius, done)
        drawing.saveSvg(Path(filename))

    def draw_frames(self, folder_name):
        for frame in range(len(self.paths[0])):
            agents = [self.paths[agent][frame] for agent in self.paths]
            done = [self.done_statuses[agent][frame] for agent in self.paths]
            targets = [self.targets[t] for a, t in enumerate(self.targets)]
            self.draw_frame(self.obstacles, targets=targets, filename=Path(folder_name) / f'{str(frame).zfill(3)}.svg',
                            agents=agents, obs_radius=self.obs_radius, done=done)


class AnimationMonitor(gym.Wrapper):
    """
    Saves svg animation renders on reset.
    """

    def __init__(self, env, directory, force=False, use_seed_name=False, min_save_length=None):
        super().__init__(env)
        self._drawer = None
        self._directory = directory

        self._force = force
        self.episode_id = 0
        self.use_seed_name = use_seed_name
        self.min_save_length = min_save_length

    def step(self, action):
        """
        Stores agents positions and done statuses on every step.
        """
        obs, reward, done, info = self.env.step(action)

        grid: Grid = self.env.grid
        self._drawer.store(positions=grid.positions_xy, done_statuses=[done])
        return obs, reward, done, info

    def reset(self, **kwargs):
        """
        Renders svg files in self._directory folder.
        """

        if self._drawer:
            if not self.min_save_length or self.min_save_length <= max(
                    [len(self._drawer.paths[agent]) for agent in self._drawer.paths]):
                drawing = self._drawer.draw()
                if self.use_seed_name:
                    filename = str(self.seed_).zfill(4) + '.svg'
                else:
                    filename = str(self.episode_id).zfill(5) + '.svg'

                path = Path(self._directory) / Path(filename)
                Path(self._directory).mkdir(parents=True, exist_ok=True)
                drawing.saveSvg(path)

        result = self.env.reset(**kwargs)
        grid: Grid = self.env.grid
        config: GridConfig = self.env.config
        targets = {index: target for index, target in enumerate(grid.finishes_xy)}
        self._drawer = MarlSvgVisualization(grid.obstacles, config.num_agents, targets, config.obs_radius)
        self._drawer.store(positions=grid.positions_xy, done_statuses=[False for _ in range(config.num_agents)])

        self.episode_id += 1
        return result



import codecs
import os
import re

from setuptools import setup, find_packages

cur_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cur_dir, 'README.md'), 'rb') as f:
    lines = [x.decode('utf-8') for x in f.readlines()]
    lines = ''.join([re.sub('^<.*>\n$', '', x) for x in lines])
    long_description = lines


def read(*parts):
    with codecs.open(os.path.join(cur_dir, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


setup(
    name='pogema',
    author='Alexey Skrynnik',
    license='MIT',
    version=find_version("pogema", "__init__.py"),
    description='Partially Observable Grid Environment for Multi-Agent RL',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Tviskaron/pogema',
    install_requires=[
        "gym>=0.18.3",
        "numpy>=1.19.2",
        "pydantic>=1.8.2",
        "drawSvg>=1.8.2",
    ],
    extras_require={
        "numba": ["numba>=0.53.1", ]
    },
    package_dir={'': './'},
    packages=find_packages(where='./', include='pogema*'),
    include_package_data=True,
    python_requires='>=3.6',
)
